﻿/* CompressionRefrigerator.cs
 * 
 * Copyright (C) 2015 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

using System;

using Popolo.ThermophysicalProperty;

namespace Popolo.HVAC
{
  /// <summary>インバータターボ冷凍機クラス</summary>
  public class CentrifugalInverterChiller
  {

    #region 定数宣言

    /// <summary>絶対温度と摂氏との変換定数</summary>
    private const double CONVERT_C_TO_K = 273.15;

    /// <summary>水の定圧比熱[kJ/(kgK)]</summary>
    private const double WATER_SPECIFIC_HEAT = 4.186;

    #endregion

    #region プロパティ・インスタンス変数

    /// <summary>機器特性係数</summary>
    private readonly double[] coef;

    /// <summary>蒸発器の熱伝達率[kW/K]</summary>
    private double evaporatorHeatTransferCoefficient;

    /// <summary>凝縮器の熱伝達率[kW/K]</summary>
    private double condenserHeatTransferCoefficient;

    /// <summary>定格断熱ヘッド[kW]</summary>
    private double nominalHead;

    /// <summary>定格冷媒流量[m3/s]</summary>
    private double nominalFlowVolume;
    
    /// <summary>冷水出口温度[C]を取得する</summary>
    public double ChilledWaterOutletTemperature { get; private set; }

    /// <summary>冷水出口温度設定値[C]を設定・取得する</summary>
    public double ChilledWaterOutletSetPointTemperature { get; set; }

    /// <summary>冷水入口温度[C]を取得する</summary>
    public double ChilledWaterInletTemperature { get; private set; }

    /// <summary>冷却水出口温度[C]を取得する</summary>
    public double CoolingWaterOutletTemperature { get; private set; }

    /// <summary>冷却水入口温度[C]を取得する</summary>
    public double CoolingWaterInletTemperature { get; private set; }

    /// <summary>冷水流量[kg/s]を取得する</summary>
    public double ChilledWaterFlowRate { get; private set; }

    /// <summary>冷却水流量[kg/s]を取得する</summary>
    public double CoolingWaterFlowRate { get; private set; }

    /// <summary>定格冷凍能力[kW]を取得する</summary>
    public double NominalCapacity { get; private set; }

    /// <summary>定格入力[kW]を取得する</summary>
    public double NominalInput { get; private set; }

    /// <summary>定格COP[-]を取得する</summary>
    public double NominalCOP { get { return NominalCapacity / NominalInput; } }

    /// <summary>容量制御下限値[-]を取得する</summary>
    public double MinimumPartialLoadRatio { get; private set; }

    /// <summary>消費電力[kW]を取得する</summary>
    public double ElectricConsumption { get; private set; }

    /// <summary>冷凍能力[kW]を取得する</summary>
    public double CoolingLoad { get; private set; }

    /// <summary>過熱度[K]を取得する</summary>
    public double SuperHeatDegree { get; private set; }

    /// <summary>COP[-]を取得する</summary>
    public double COP { get { return CoolingLoad / ElectricConsumption; } }

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタ</summary>
    /// <param name="nominalInput">定格入力[kW]</param>
    /// <param name="minimumPartialLoadRatio">容量制御下限値[-]</param>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="chilledWaterOutletTemperature">冷水出口温度[C]</param>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="chilledWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="coolingWaterFlowRate">冷却水質量流量[kg/s]</param>
    public CentrifugalInverterChiller
      (double nominalInput, double minimumPartialLoadRatio, 
      double chilledWaterInletTemperature, double chilledWaterOutletTemperature,
      double coolingWaterInletTemperature, double chilledWaterFlowRate,
      double coolingWaterFlowRate)
    {
      this.ChilledWaterOutletSetPointTemperature = chilledWaterOutletTemperature;
      this.CoolingWaterInletTemperature = coolingWaterInletTemperature;
      this.ChilledWaterInletTemperature = chilledWaterInletTemperature;
      this.CoolingWaterFlowRate = coolingWaterFlowRate;
      this.ChilledWaterFlowRate = chilledWaterFlowRate;
      this.NominalCapacity = chilledWaterFlowRate * WATER_SPECIFIC_HEAT *
        (chilledWaterInletTemperature - chilledWaterOutletTemperature);
      this.NominalInput = nominalInput;
      this.MinimumPartialLoadRatio = minimumPartialLoadRatio;

      //特性係数初期化//汎用モデルで固定
      coef = new double[] { 0.8553, 0.0590, 1.8515, -3.9452, 3.4178, -0.0945, 0.0438 };

      //蒸発器の熱伝達率[kW/K]を計算
      double tEvp = ChilledWaterOutletSetPointTemperature - 2;
      double dt1 = ChilledWaterInletTemperature- tEvp;
      double dt2 = ChilledWaterOutletSetPointTemperature - tEvp;
      double lmtd = (dt1 - dt2) / Math.Log(dt1 / dt2);
      evaporatorHeatTransferCoefficient = NominalCapacity / lmtd;

      //凝縮器の熱伝達率[kW/K]を計算
      double qcd = NominalCapacity + NominalInput;
      CoolingWaterOutletTemperature = CoolingWaterInletTemperature +
        qcd / (WATER_SPECIFIC_HEAT * coolingWaterFlowRate);
      double tCnd = CoolingWaterOutletTemperature + 1;
      dt1 = tCnd - CoolingWaterInletTemperature;
      dt2 = tCnd - CoolingWaterOutletTemperature;
      lmtd = (dt1 - dt2) / Math.Log(dt1 / dt2);
      condenserHeatTransferCoefficient = qcd / lmtd;

      //定格断熱圧縮仕事[kW]と冷媒流量[m3/s]を計算
      double mccd = CoolingWaterFlowRate * WATER_SPECIFIC_HEAT;
      double mcch = ChilledWaterFlowRate * WATER_SPECIFIC_HEAT;
      getHeadAndFlowVolume(qcd, NominalCapacity,
        CoolingWaterInletTemperature, ChilledWaterInletTemperature,
        condenserHeatTransferCoefficient, evaporatorHeatTransferCoefficient,
        mccd, mcch, coef, out nominalHead, out nominalFlowVolume);
    }

    #endregion    

    #region publicメソッド

    /// <summary>状態を更新する</summary>
    /// <param name="coolingWaterInletTemperature">冷却水入口温度[C]</param>
    /// <param name="chilledWaterInletTemperature">冷水入口温度[C]</param>
    /// <param name="coolingWaterFlowRate">冷水質量流量[kg/s]</param>
    /// <param name="chilledWaterFlowRate">冷却水質量流量[kg/s]</param>
    public void Update
      (double coolingWaterInletTemperature, double chilledWaterInletTemperature,
      double coolingWaterFlowRate, double chilledWaterFlowRate)
    {
      //状態値を保存
      this.ChilledWaterInletTemperature = chilledWaterInletTemperature;
      this.CoolingWaterInletTemperature = coolingWaterInletTemperature;
      this.ChilledWaterFlowRate = chilledWaterFlowRate;
      this.CoolingWaterFlowRate = coolingWaterFlowRate;

      //熱容量流量[kW/K]を計算
      double mccd = CoolingWaterFlowRate * WATER_SPECIFIC_HEAT;
      double mcch = ChilledWaterFlowRate * WATER_SPECIFIC_HEAT;

      //必要能力[kW]を計算
      CoolingLoad = mcch *
        (ChilledWaterInletTemperature - ChilledWaterOutletSetPointTemperature);
      double partialLoad = 
        Math.Max(MinimumPartialLoadRatio, CoolingLoad / NominalCapacity);

      //過負荷判定
      double head, volume;
      getHeadAndFlowVolume(CoolingLoad + NominalInput, CoolingLoad,
        CoolingWaterInletTemperature, ChilledWaterInletTemperature,
        condenserHeatTransferCoefficient, evaporatorHeatTransferCoefficient,
        mccd, mcch, coef, out head, out volume);
      ElectricConsumption = getElectricity(head, volume, partialLoad);

      const double TOL = 0.001;
      const double DELTA = 0.0001;

      //最大能力範囲内:消費電力を収束計算
      if (ElectricConsumption < NominalInput)
      {
        double err1 = eFnc1(ElectricConsumption, mccd, mcch, partialLoad);
        while (TOL < Math.Abs(err1))
        {
          double err2 = eFnc1
            (ElectricConsumption + DELTA, mccd, mcch, partialLoad);
          ElectricConsumption -= err1 * DELTA / (err2 - err1);
          err1 = eFnc1(ElectricConsumption, mccd, mcch, partialLoad);
        }
        ChilledWaterOutletTemperature = ChilledWaterOutletSetPointTemperature;
      }
      //過負荷//冷却能力を収束計算
      else
      {
        CoolingLoad = NominalCapacity;
        double err1 = eFnc2(CoolingLoad, mccd, mcch, partialLoad);
        while (TOL < Math.Abs(err1))
        {
          double err2 = eFnc2
            (CoolingLoad + DELTA, mccd, mcch, partialLoad);
          CoolingLoad -= err1 * DELTA / (err2 - err1);
          err1 = eFnc2(CoolingLoad, mccd, mcch, partialLoad);
        }
        ChilledWaterOutletTemperature = 
          ChilledWaterInletTemperature - CoolingLoad / mcch;
      }
    }
    
    #endregion

    #region privateメソッド

    /// <summary>誤差関数</summary>
    /// <param name="electricity">消費電力[kW]</param>
    /// <param name="mccd">冷却水熱容量流量[kW/K]</param>
    /// <param name="mcch">冷水熱容量流量[kW/K]</param>
    /// <param name="partialLoad">負荷率[-]</param>
    /// <returns>誤差</returns>
    private double eFnc1
      (double electricity, double mccd, double mcch, double partialLoad)
    {
      double head, volume;
      getHeadAndFlowVolume(CoolingLoad + electricity, CoolingLoad,
        CoolingWaterInletTemperature, ChilledWaterInletTemperature,
        condenserHeatTransferCoefficient, evaporatorHeatTransferCoefficient,
        mccd, mcch, coef, out head, out volume);
      return getElectricity(head, volume, partialLoad) - electricity;
    }

    /// <summary>誤差関数</summary>
    /// <param name="coolingLoad">冷房負荷[kW]</param>
    /// <param name="mccd">冷却水熱容量流量[kW/K]</param>
    /// <param name="mcch">冷水熱容量流量[kW/K]</param>
    /// <param name="partialLoad">負荷率[-]</param>
    /// <returns>誤差</returns>
    private double eFnc2
      (double coolingLoad, double mccd, double mcch, double partialLoad)
    {
      double head, volume;
      getHeadAndFlowVolume(coolingLoad + NominalInput, coolingLoad,
        CoolingWaterInletTemperature, ChilledWaterInletTemperature,
        condenserHeatTransferCoefficient, evaporatorHeatTransferCoefficient,
        mccd, mcch, coef, out head, out volume);
      return getElectricity(head, volume, partialLoad) - NominalInput;
    }

    /// <summary>消費電力[kW]を計算する</summary>
    /// <param name="head">断熱圧縮仕事[kW]</param>
    /// <param name="volume">体積流量[m3/s]</param>
    /// <param name="partialLoad">負荷率[-]</param>
    /// <returns>消費電力[kW]</returns>
    private double getElectricity
      (double head, double volume, double partialLoad)
    {
      double rvR = volume /
        ((Math.Sqrt(head / nominalHead) * coef[0] + coef[1]) * nominalFlowVolume);
      return head * (coef[4] + rvR * (coef[3] + rvR * coef[2])) +
        NominalInput * (partialLoad * coef[5] + coef[6]);
    }

    /// <summary>断熱圧縮仕事[kW]と体積流量[m3/s]を計算する</summary>
    /// <param name="qcd">凝縮器熱交換量[kW]</param>
    /// <param name="qch">蒸発器熱交換量[kW]</param>
    /// <param name="tcdi">冷却水入口温度[C]</param>
    /// <param name="tchi">冷水入口温度[C]</param>
    /// <param name="kacd">凝縮器熱通過率[kW/K]</param>
    /// <param name="kach">蒸発器熱通過率[kW/K]</param>
    /// <param name="mccd">凝縮器熱容量流量[kW/K]</param>
    /// <param name="mcch">蒸発器熱容量流量[kW/K]</param>
    /// <param name="coefs">特性係数</param>
    /// <param name="head">出力:断熱圧縮仕事[kW]</param>
    /// <param name="volume">出力:体積流量[m3/s]</param>
    private static void getHeadAndFlowVolume
      (double qcd, double qch, double tcdi, double tchi, double kacd, double kach,
      double mccd, double mcch, double[] coefs, out double head, out double volume)
    {
      //凝縮温度・圧力を計算
      double tCnd = tcdi + qcd / ((1 - Math.Exp(-kacd / mccd)) * mccd);
      double dlCnd, dvCnd, pCnd;
      Refrigerant.GetSaturatedPropertyFromTemperature
        (tCnd + 273.15, Refrigerant.Fluid.R134a, out dlCnd, out dvCnd, out pCnd);
      double hoCnd = Refrigerant.GetEnthalpyFromTemperatureAndDensity
        (tCnd + 273.15, dlCnd, Refrigerant.Fluid.R134a);

      //蒸発温度・圧力を計算
      double tEvp = tchi - qch / ((1 - Math.Exp(-kach / mcch)) * mcch);
      double dlEvp, dvEvp, pEvp;
      Refrigerant.GetSaturatedPropertyFromTemperature
        (tEvp + 273.15, Refrigerant.Fluid.R134a, out dlEvp, out dvEvp, out pEvp);
      double hiCmp1 = Refrigerant.GetEnthalpyFromTemperatureAndDensity
        (tEvp + 273.15, dvEvp, Refrigerant.Fluid.R134a);

      //中間圧力を計算
      double pMid = Math.Sqrt(pEvp * pCnd);
      double dlMid, dvMid, tMid;
      Refrigerant.GetSaturatedPropertyFromPressure
        (pMid, Refrigerant.Fluid.R134a, out dlMid, out dvMid, out tMid);
      double hsvMid = Refrigerant.GetEnthalpyFromTemperatureAndDensity
        (tMid, dvMid, Refrigerant.Fluid.R134a);

      //蒸発器冷媒流量を計算
      double hiEvp = Refrigerant.GetEnthalpyFromTemperatureAndDensity
        (tMid, dlMid, Refrigerant.Fluid.R134a);
      double mRevp = qch / (hiCmp1 - hiEvp);
      volume = mRevp / dvEvp;

      //1段目羽根車の断熱ヘッド[kW]と出口比エンタルピー[kJ/kg]を計算
      double kappa = Refrigerant.GetSpecificHeatRatioFromTemperatureAndDensity
        (tEvp + 273.15, dvEvp, Refrigerant.Fluid.R134a);
      kappa = (kappa - 1) / kappa;
      double head1 = volume / kappa * pEvp * (Math.Pow(pMid / pEvp, kappa) - 1);
      double hoCmp1 = hiCmp1 + head1 / mRevp;

      //2段目羽根車断熱ヘッド[kW]を計算
      double mRe = (hoCnd - hiEvp) / (hsvMid - hoCnd) * mRevp;
      double hiCmp2 = (hoCmp1 * mRevp + hsvMid * mRe) / (mRevp + mRe);
      double tiCmp2, diCmp2, siCmp2, uiCmp2;
      Refrigerant.GetStateFromPressureAndEnthalpy
        (pMid, hiCmp2, Refrigerant.Fluid.R134a,
        out tiCmp2, out diCmp2, out siCmp2, out uiCmp2);
      double vRCmp2 = (mRevp + mRe) / diCmp2;
      kappa = Refrigerant.GetSpecificHeatRatioFromTemperatureAndDensity
        (tiCmp2 + 273.15, diCmp2, Refrigerant.Fluid.R134a);
      kappa = (kappa - 1) / kappa;
      double head2 = vRCmp2 / kappa * pMid * (Math.Pow(pCnd / pMid, kappa) - 1);

      //圧縮機入力[kW]の計算
      head = head1 + head2;
    }

    #endregion

  }
}
