﻿/* Refrigerant.cs
 * 
 * Copyright (C) 2013 E.Togashi
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

using System;

namespace Popolo.ThermophysicalProperty
{
  /// <summary>冷媒物性値計算クラス</summary>
  public static class Refrigerant
  {

    #region 定数宣言

    /// <summary>絶対温度と摂氏との変換定数</summary>
    private const double CONVERT_C_TO_K = 273.15;

    #endregion

    #region 列挙型定義

    /// <summary>冷媒の種類</summary>
    public enum Fluid
    {
      /// <summary>R410A</summary>
      R410A,
      /// <summary>R32</summary>
      R32,
      /// <summary>R134a</summary>
      R134a
    }

    /// <summary>冷媒の状態</summary>
    public enum Phase
    {
      /// <summary>気相</summary>
      Vapor,
      /// <summary>液相</summary>
      Liquid,
      /// <summary>気液平衡</summary>
      Equilibrium
    }

    #endregion

    #region クラス変数定義

    /// <summary>現在の冷媒種類を保持する</summary>
    private static Fluid currentFluid = Fluid.R410A;

    #endregion

    #region 近似係数

    #region 近似係数

    /// <summary>近似係数mの数</summary>
    private static int m_number = 5;

    /// <summary>近似係数nの数</summary>
    private static int n_number = 7;

    /// <summary>臨界温度[K]</summary>
    private static double criticalTemperature = 78.105 + CONVERT_C_TO_K;

    /// <summary>臨界密度[kg/m3]</summary>
    private static double criticalDensity = 424.0;

    /// <summary>臨界圧力[kPa]</summary>
    private static double criticalPressure = 5782;

    /// <summary>ガス定数</summary>
    private static double gasConstant = 8.3144621 / 52.024;

    /// <summary>基準状態の温度[K]</summary>
    private static double refTemperature = 0 + CONVERT_C_TO_K;

    /// <summary>基準状態の密度[kg/m3]</summary>
    private static double refDensity = 1055.3;

    /// <summary>基準状態のエンタルピー[kJ/kg]</summary>
    private static double refEnthalpy = 200;

    /// <summary>基準状態のエントロピー[kJ/kgK]</summary>
    private static double refEntropy = 1.0;

    /// <summary>PVT近似係数</summary>
    private static double[,] alpha;

    /// <summary>定圧比熱[kJ/kg]近似係数</summary>
    private static double[] ccp = new double[] { 4.0186023E+00, -3.1370881E-01, 6.8796834E-01, 2.6831619E+00, -1.3934091E+00 };

    /// <summary>飽和圧力[kPa]の初期値推測用近似係数</summary>
    private static double[] cps = new double[] { 85607, -161288, 104677, -23310 };

    /// <summary>飽和温度[K]の初期値推測用近似係数</summary>
    private static double[] cts = new double[] { 439, -594, 361, 232 };

    #endregion

    #region 近似係数（R32）

    /// <summary>PVT近似係数</summary>
    private static double[] alpha_R32 = new double[] { 9.7665906E+03, 3.4609053E+04, -4.1024673E+04, 1.1966225E+04, -1.6994460E+02, 1.9091076E+02, 1.0932188E+03, -8.0842415E+03, 1.4466692E+04, -1.0718168E+04, 3.5651716E+03, -4.4110895E+02, -1.3085150E+05, -1.8447255E+05, 2.9538016E+05, -7.9536583E+04, -1.0279583E+04, 3.3629135E+03, 1.2591969E+05, 3.3280998E+05, -5.4534004E+05, 1.9771955E+05, -7.5414922E+03, -3.2633069E+03, -4.3690554E+04, -1.4724183E+05, 2.7062311E+05, -1.2341107E+05, 1.6622311E+04, 0.0000000E+00 };

    /// <summary>定圧比熱[kJ/kg]近似係数</summary>
    private static double[] ccp_R32 = new double[] { 4.0186023E+00, -3.1370881E-01, 6.8796834E-01, 2.6831619E+00, -1.3934091E+00 };

    /// <summary>飽和圧力[kPa]の初期値推測用近似係数</summary>
    private static double[] cps_R32 = new double[] { 102795, -204886, 141476, -33645 };

    /// <summary>飽和温度[K]の初期値推測用近似係数</summary>
    private static double[] cts_R32 = new double[] { 148, -298, 269, 241 };

    #endregion

    #region 近似係数（R410A）

    /// <summary>PVT近似係数</summary>
    private static double[] alpha_R410A = new double[] { 6.7196006E+03, 3.1935694E+04, -3.2418379E+04, 8.5482532E-01, 8.1295227E+03, -1.7327422E+03, 8.6392377E+02, -5.7171597E+03, 6.8653458E+03, -3.3190647E+03, 7.1308696E+02, -5.6028267E+01, -8.5841472E+04, -1.9218673E+05, 3.2004848E+05, -1.2678673E+05, 3.6577226E+03, 3.5479124E+03, 8.2190468E+04, 3.1572566E+05, -5.5079066E+05, 2.6367950E+05, -3.7740069E+04, -4.9244792E+02, -2.8629986E+04, -1.3682605E+05, 2.5924131E+05, -1.3997914E+05, 2.7545424E+04, -1.4185289E+03 };

    /// <summary>定圧比熱[kJ/kg]近似係数</summary>
    private static double[] ccp_R410A = new double[] { 3.8200427E+00, 2.5486066E+00, 1.0799339E+00, 9.8471295E-01, -7.6910990E-01 };

    /// <summary>飽和圧力[kPa]の初期値推測用近似係数</summary>
    private static double[] cps_R410A = new double[] { 86584, -171854, 118095, -27955 };

    /// <summary>飽和温度[K]の初期値推測用近似係数</summary>
    private static double[] cts_R410A = new double[] { 151, -292, 260, 238 };

    #endregion

    #region 近似係数（R134a）

    /// <summary>PVT近似係数</summary>
    private static double[] alpha_R134a = new double[] { 9.6110418E+03, 2.9590018E+04, -2.2904579E+04, 2.7602291E+03, 6.3436858E+02, -5.9229642E+01, 1.1017426E+03, -1.2925788E+03, 5.1927422E+02, -6.9885390E+01, -7.5524176E+04, -2.8006887E+05, 3.3388181E+05, -1.0398404E+05, 9.7152916E+03, 6.4037701E+04, 4.5332427E+05, -5.7147500E+05, 1.9611019E+05, -2.0375143E+04, -1.9195915E+04, -1.9499885E+05, 2.6634260E+05, -1.0132763E+05, 1.1874162E+04 };

    /// <summary>定圧比熱[kJ/kg]近似係数</summary>
    private static double[] ccp_R134a = new double[] { 1.9747230E-01, 2.6221924E+01, -3.8424120E+01, 3.8198553E+01, -1.4361012E+01 };

    /// <summary>飽和圧力[kPa]の初期値推測用近似係数</summary>
    private static double[] cps_R134a = new double[] { 56756, -104220, 65405, -13992 };

    /// <summary>飽和温度[K]の初期値推測用近似係数</summary>
    private static double[] cts_R134a = new double[] { 1355, -1252, 513, 242 };

    #endregion

    #endregion

    #region コンストラクタ

    /// <summary>コンストラクタ</summary>
    static Refrigerant()
    {
      //強制的に初期化
      currentFluid = Fluid.R410A;
      initRefrigerant(Fluid.R32);
    }

    /// <summary>近似係数を初期化する</summary>
    /// <param name="fluid">冷媒種類</param>
    private static void initRefrigerant(Refrigerant.Fluid fluid)
    {
      //指定冷媒で初期化済の場合は終了
      if (currentFluid == fluid) return;
      int index;

      switch (fluid)
      {
        case Fluid.R410A:
          m_number = 5;
          n_number = 8;
          criticalTemperature = 71.358 + CONVERT_C_TO_K;
          criticalDensity = 459.53;
          criticalPressure = 4902.6;
          gasConstant = 8.3144621 / 72.585;
          refTemperature = 0 + CONVERT_C_TO_K;
          refDensity = 1170;
          refEnthalpy = 200;
          refEntropy = 1.0;
          ccp = ccp_R410A;
          cps = cps_R410A;
          cts = cts_R410A;
          alpha = new double[m_number, n_number - 2];
          index = 0;
          for (int m = 0; m < alpha.GetLength(0); m++)
          {
            for (int n = 0; n < alpha.GetLength(1); n++)
            {
              alpha[m, n] = alpha_R410A[index];
              index++;
            }
          }
          break;
        case Fluid.R32:
          m_number = 5;
          n_number = 8;
          criticalTemperature = 78.105 + CONVERT_C_TO_K;
          criticalDensity = 424.0;
          criticalPressure = 5782;
          gasConstant = 8.3144621 / 52.024;
          refTemperature = 0 + CONVERT_C_TO_K;
          refDensity = 1055.3;
          refEnthalpy = 200;
          refEntropy = 1.0;
          ccp = ccp_R32;
          cps = cps_R32;
          cts = cts_R32;
          alpha = new double[m_number, n_number - 2];
          index = 0;
          for (int m = 0; m < alpha.GetLength(0); m++)
          {
            for (int n = 0; n < alpha.GetLength(1); n++)
            {
              alpha[m, n] = alpha_R32[index];
              index++;
            }
          }
          break;
        case Fluid.R134a:
          m_number = 5;
          n_number = 7;
          criticalTemperature = 101.06 + CONVERT_C_TO_K;
          criticalDensity = 511.9;
          criticalPressure = 4059.3;
          gasConstant = 8.3144621 / 102.03;
          refTemperature = 0 + CONVERT_C_TO_K;
          refDensity = 1294.8;
          refEnthalpy = 200;
          refEntropy = 1.0;
          ccp = ccp_R134a;
          cps = cps_R134a;
          cts = cts_R134a;
          alpha = new double[m_number, n_number - 2];
          index = 0;
          for (int m = 0; m < alpha.GetLength(0); m++)
          {
            for (int n = 0; n < alpha.GetLength(1); n++)
            {
              alpha[m, n] = alpha_R134a[index];
              index++;
            }
          }
          break;
        default:

          break;
      }
      //現在の冷媒種類を更新
      currentFluid = fluid;
    }

    #endregion

    #region 温度[K]および密度[kg/m3]による物性値計算処理

    /// <summary>温度[K]および密度[kg/m3]から圧力[kPa]を計算する</summary>
    /// <param name="temperature">温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>圧力[kPa]</returns>
    public static double GetPressureFromTemperatureAndDensity
        (double temperature, double density, Fluid fluid)
    {
      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      double tau = criticalTemperature / temperature;
      double rho = density / criticalDensity;

      double pressure = 0;
      for (int m = alpha.GetLength(0) - 1; 0 <= m; m--)
      {
        double buff = 0;
        for (int n = alpha.GetLength(1) - 1; 0 <= n; n--)
        {
          buff = buff * rho + alpha[m, n];
        }
        pressure = pressure * tau + buff;
      }

      return pressure * rho * rho + density * gasConstant * temperature;
    }

    /// <summary>温度[K]および密度[kg/m3]から比エンタルピー[kJ/kg]を計算する</summary>
    /// <param name="temperature">温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>比エンタルピー[kJ/kg]</returns>
    public static double GetEnthalpyFromTemperatureAndDensity
        (double temperature, double density, Fluid fluid)
    {
      double h = getResidualEnthalpy(temperature, density, fluid) -
          getResidualEnthalpy(refTemperature, refDensity, fluid);
      h += getIntegralCp0(temperature);
      return h + refEnthalpy;
    }

    /// <summary>温度[K]および密度[kg/m3]から比エントロピー[kJ/kgK]を計算する</summary>
    /// <param name="temperature">温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>比エントロピー[kJ/kgK]</returns>
    public static double GetEntropyFromTemperatureAndDensity
        (double temperature, double density, Fluid fluid)
    {
      double s = getResidualEntropy(temperature, density, fluid) -
          getResidualEntropy(refTemperature, refDensity, fluid);
      s += gasConstant * Math.Log(refDensity / density);
      s += getIntegralCp0T(temperature) - gasConstant * Math.Log(temperature / refTemperature);
      return s + refEntropy;
    }

    /// <summary>温度[K]および密度[kg/m3]から内部エネルギー[kJ/kg]を計算する</summary>
    /// <param name="temperature">温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>内部エネルギー[kJ/kg]</returns>
    public static double GetInternalEnergyFromTemperatureAndDensity
        (double temperature, double density, Fluid fluid)
    {
      double p = GetPressureFromTemperatureAndDensity(temperature, density, fluid);
      return GetEnthalpyFromTemperatureAndDensity
          (temperature, density, fluid) - p / density;
    }

    /// <summary>絶対温度[K]と密度[kg/m3]から定積比熱[kJ/kgK]を計算する</summary>
    /// <param name="temperature">絶対温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>定積比熱[kJ/kgK]</returns>
    public static double GetIsovolumetricSpecificHeatFromTemperatureAndDensity
        (double temperature, double density, Fluid fluid)
    {
      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      double tau = criticalTemperature / temperature;
      double rho = density / criticalDensity;

      double cv = 0;
      for (int m = alpha.GetLength(0) - 1; 2 <= m; m--)
      {
        double buff = 0;
        for (int n = alpha.GetLength(1) + 1; 2 <= n; n--)
        {
          buff = buff * rho + alpha[m, n - 2] * m * (m + 1) / (1d - n);
        }
        cv = cv * tau + buff;
      }
      cv *= rho * tau * tau;

      return cv / (criticalDensity * temperature) + getIsovolumetricHeatCapacityOfIdealGas(temperature);
    }

    /// <summary>絶対温度[K]と密度[kg/m3]から定圧比熱[kJ/kgK]を計算する</summary>
    /// <param name="temperature">絶対温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>定圧比熱[kJ/kgK]</returns>
    public static double GetIsobaricSpecificHeatFromTemperatureAndDensity
        (double temperature, double density, Fluid fluid)
    {
      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      double tau = criticalTemperature / temperature;
      double rho = density / criticalDensity;

      double cp1 = 0;
      double cp2 = 0;
      for (int m = alpha.GetLength(0) - 1; 0 <= m; m--)
      {
        double buff1 = 0;
        double buff2 = 0;
        for (int n = alpha.GetLength(1) + 1; 2 <= n; n--)
        {
          if (m != 0) buff1 = buff1 * rho + alpha[m, n - 2] * m;
          buff2 = buff2 * rho + alpha[m, n - 2] * n;
        }
        if (m != 0) cp1 = cp1 * tau + buff1;
        cp2 = cp2 * tau + buff2;
      }
      cp1 = density * gasConstant - cp1 * rho * rho * tau / temperature;
      cp2 = gasConstant * temperature + cp2 * rho / criticalDensity;

      double bf = cp1 / density;
      return GetIsovolumetricSpecificHeatFromTemperatureAndDensity(temperature, density, fluid)
        + temperature * bf * bf / cp2;
    }

    /// <summary>絶対温度[K]と密度[kg/m3]から比熱比[-]を計算する</summary>
    /// <param name="temperature">絶対温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>比熱比[-]</returns>
    public static double GetSpecificHeatRatioFromTemperatureAndDensity
       (double temperature, double density, Fluid fluid)
    {
      double cv = GetIsovolumetricSpecificHeatFromTemperatureAndDensity
        (temperature, density, fluid);
      double cp = GetIsobaricSpecificHeatFromTemperatureAndDensity
        (temperature, density, fluid);
      return cp / cv;
    }

    /// <summary>温度[K]および密度[kg/m3]から圧力[kPa]の密度微分値[kPa/(kg/m3)]を計算する</summary>
    /// <param name="temperature">温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>圧力[kPa]の密度微分値[kPa/(kg/m3)]</returns>
    private static double getDensityDerivativeOfPressure
        (double temperature, double density, Fluid fluid)
    {
      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      double tau = criticalTemperature / temperature;
      double rho = density / criticalDensity;

      double dPdRho = 0;
      for (int m = alpha.GetLength(0) - 1; 0 <= m; m--)
      {
        double buff = 0;
        for (int n = alpha.GetLength(1) + 1; 2 <= n; n--)
        {
          buff = buff * rho + n * alpha[m, n - 2];
        }
        dPdRho = dPdRho * tau + buff;
      }
      dPdRho *= rho;

      return dPdRho / criticalDensity + gasConstant * temperature;
    }

    #endregion

    #region エンタルピー[kJ/kg]および圧力[kPa]による物性値計算処理

    /// <summary>圧力[kPa]と比エンタルピー[kJ/kg]から冷媒状態を計算する</summary>
    /// <param name="pressure">圧力[kPa]</param>
    /// <param name="enthalpy">比エンタルピー[kJ/kg]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <param name="temperature">絶対温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="entropy">比エントロピー[kJ/kgK]</param>
    /// <param name="internalEnergy">比内部エネルギー[kJ/kg]</param>
    public static void GetStateFromPressureAndEnthalpy
        (double pressure, double enthalpy, Fluid fluid,
        out double temperature, out double density, out double entropy, out double internalEnergy)
    {
      const double DELTA = 0.001;
      const double ERR_TOL = 0.01;
      const int MAX_ITER = 10;

      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      //冷媒の状態を判定
      double rhoL, rhoV, tSat;
      GetSaturatedPropertyFromPressure(pressure, fluid, out rhoL, out rhoV, out tSat);
      Phase phase;
      double hl = GetEnthalpyFromTemperatureAndDensity(tSat, rhoL, fluid);
      double hv = GetEnthalpyFromTemperatureAndDensity(tSat, rhoV, fluid);
      if (enthalpy < hl) phase = Phase.Liquid;
      else
      {
        if (hv < enthalpy) phase = Phase.Vapor;
        else phase = Phase.Equilibrium;
      }

      //冷媒の状態に応じて密度初期値を設定
      if (phase == Phase.Equilibrium)
      {
        temperature = tSat;
        double vRate = (enthalpy - hl) / (hv - hl);
        double lRate = 1 - vRate;
        //密度
        density = 1 / (vRate / rhoV + lRate / rhoL);
        //比エントロピー
        double sL = GetEntropyFromTemperatureAndDensity(tSat, rhoL, fluid);
        double sV = GetEntropyFromTemperatureAndDensity(tSat, rhoV, fluid);
        entropy = sL * lRate + sV * vRate;
        //比内部エネルギー
        double uL = GetInternalEnergyFromTemperatureAndDensity(tSat, rhoL, fluid);
        double uV = GetInternalEnergyFromTemperatureAndDensity(tSat, rhoV, fluid);
        internalEnergy = uL * lRate + uV * vRate;
        return; //二相域の場合には温度確定のため収束計算不要
      }
      else if (phase == Phase.Liquid)
      {
        temperature = tSat - 3;
        density = rhoL;
      }
      else
      {
        temperature = tSat + 3;
        density = rhoV;
      }

      //ニュートン法で温度を収束計算
      double err = enthalpy - GetEnthalpyFromTemperatureAndDensity(temperature, density, fluid);
      int iter = 0;
      while (ERR_TOL < Math.Abs(err))
      {
        if (MAX_ITER < iter) throw new Exception("Iteration Error:GetStateFromPressureAndEnthalpy");
        iter++;

        //誤差関数の数値微分を計算
        getDensityFromPressureAndTemperature(pressure, temperature + DELTA, fluid, ref density);
        double err2 = enthalpy - GetEnthalpyFromTemperatureAndDensity(temperature + DELTA, density, fluid);
        double dEdT = (err2 - err) / DELTA;

        temperature -= err / dEdT;
        getDensityFromPressureAndTemperature(pressure, temperature, fluid, ref density);
        err = enthalpy - GetEnthalpyFromTemperatureAndDensity(temperature, density, fluid);
      }
      //比エントロピー、比内部エネルギー
      entropy = GetEntropyFromTemperatureAndDensity(temperature, density, fluid);
      internalEnergy = GetInternalEnergyFromTemperatureAndDensity(temperature, density, fluid);
    }

    #endregion

    #region エントロピー[kJ/kgK]および圧力[kPa]による物性値計算処理

    /// <summary>圧力[kPa]と比エントロピー[kJ/kg]から他の物性値を計算する</summary>
    /// <param name="pressure">圧力[kPa]</param>
    /// <param name="entropy">比エントロピー[kJ/kgK]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <param name="temperature">絶対温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="enthalpy">比エンタルピー[kJ/kg]</param>
    /// <param name="internalEnergy">比内部エネルギー[kJ/kg]</param>
    public static void GetStateFromPressureAndEntropy
        (double pressure, double entropy, Fluid fluid,
        out double temperature, out double density, out double enthalpy, out double internalEnergy)
    {
      const double DELTA = 0.001;
      const double ERR_TOL = 0.01;
      const int MAX_ITER = 10;

      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      //冷媒の状態を判定
      double rhoL, rhoV, tSat;
      GetSaturatedPropertyFromPressure(pressure, fluid, out rhoL, out rhoV, out tSat);
      Phase phase;
      double sl = GetEntropyFromTemperatureAndDensity(tSat, rhoL, fluid);
      double sv = GetEntropyFromTemperatureAndDensity(tSat, rhoV, fluid);
      if (entropy < sl) phase = Phase.Liquid;
      else
      {
        if (sv < entropy) phase = Phase.Vapor;
        else phase = Phase.Equilibrium;
      }

      //冷媒の状態に応じて密度初期値を設定
      if (phase == Phase.Equilibrium)
      {
        temperature = tSat;
        double vRate = (entropy - sl) / (sv - sl);
        double lRate = 1 - vRate;
        //密度
        density = 1 / (vRate / rhoV + lRate / rhoL);
        //比エンタルピー
        double hL = GetEnthalpyFromTemperatureAndDensity(tSat, rhoL, fluid);
        double hV = GetEnthalpyFromTemperatureAndDensity(tSat, rhoV, fluid);
        enthalpy = hL * lRate + hV * vRate;
        //比内部エネルギー
        double uL = GetInternalEnergyFromTemperatureAndDensity(tSat, rhoL, fluid);
        double uV = GetInternalEnergyFromTemperatureAndDensity(tSat, rhoV, fluid);
        internalEnergy = uL * lRate + uV * vRate;
        return; //二相域の場合には温度確定のため収束計算不要
      }
      else if (phase == Phase.Liquid)
      {
        temperature = tSat - 3;
        density = rhoL;
      }
      else
      {
        temperature = tSat + 3;
        density = rhoV;
      }

      //ニュートン法で温度を収束計算
      double err = entropy - GetEntropyFromTemperatureAndDensity(temperature, density, fluid);
      int iter = 0;
      while (ERR_TOL < Math.Abs(err))
      {
        if (MAX_ITER < iter) throw new Exception("Iteration Error:GetStateFromPressureAndEntropy");
        iter++;

        //誤差関数の数値微分を計算
        getDensityFromPressureAndTemperature(pressure, temperature + DELTA, fluid, ref density);
        double err2 = entropy - GetEntropyFromTemperatureAndDensity(temperature + DELTA, density, fluid);
        double dEdT = (err2 - err) / DELTA;

        temperature -= err / dEdT;
        getDensityFromPressureAndTemperature(pressure, temperature, fluid, ref density);
        err = entropy - GetEntropyFromTemperatureAndDensity(temperature, density, fluid);
      }
      //比エンタルピー、比内部エネルギー
      enthalpy = GetEnthalpyFromTemperatureAndDensity(temperature, density, fluid);
      internalEnergy = GetInternalEnergyFromTemperatureAndDensity(temperature, density, fluid);
    }

    #endregion

    #region 温度[K]および圧力[kPa]による物性値計算処理

    /// <summary>圧力[kPa]と絶対温度[K]から他の物性値を計算する</summary>
    /// <param name="pressure">圧力[kPa]</param>
    /// <param name="temperature">絶対温度[K]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <param name="entropy">比エントロピー[kJ/kgK]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="enthalpy">比エンタルピー[kJ/kg]</param>
    /// <param name="internalEnergy">比内部エネルギー[kJ/kg]</param>
    public static void GetStateFromPressureAndTemperature
        (double pressure, double temperature, Fluid fluid,
        out double entropy, out double density, out double enthalpy, out double internalEnergy)
    {
      const double DELTA = 0.001;
      const double ERR_TOL = 0.01;
      const int MAX_ITER = 10;

      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      //冷媒の状態を判定
      double rhoL, rhoV, tSat;
      GetSaturatedPropertyFromPressure(pressure, fluid, out rhoL, out rhoV, out tSat);
      Phase phase;
      if (temperature < tSat) phase = Phase.Liquid;
      else
      {
        if (tSat < temperature) phase = Phase.Vapor;
        else phase = Phase.Equilibrium;
      }

      //冷媒の状態に応じて密度初期値を設定
      if (phase == Phase.Equilibrium)
      {
        //二相域の場合には温度情報から物性は定まらない
        //飽和液の物性を出力する
        density = 1 / rhoL;
        enthalpy = GetEnthalpyFromTemperatureAndDensity(tSat, rhoL, fluid);
        internalEnergy = GetInternalEnergyFromTemperatureAndDensity(tSat, rhoL, fluid);
        entropy = GetEntropyFromTemperatureAndDensity(tSat, rhoL, fluid);
        return;
      }
      else if (phase == Phase.Liquid)
        density = rhoL + 0.1;
      else
        density = rhoV - 0.1;

      //ニュートン法で温度を収束計算
      double err = pressure - GetPressureFromTemperatureAndDensity(temperature, density, fluid);
      int iter = 0;
      while (ERR_TOL < Math.Abs(err))
      {
        if (MAX_ITER < iter) throw new Exception("Iteration Error:GetStateFromPressureAndEntropy");
        iter++;

        //誤差関数の数値微分を計算
        double err2 = pressure - GetPressureFromTemperatureAndDensity(temperature, density + DELTA, fluid);
        double dEdRHO = (err2 - err) / DELTA;

        density -= err / dEdRHO;
        err = pressure - GetPressureFromTemperatureAndDensity(temperature, density, fluid);
      }
      //比エンタルピー、比内部エネルギー
      enthalpy = GetEnthalpyFromTemperatureAndDensity(temperature, density, fluid);
      internalEnergy = GetInternalEnergyFromTemperatureAndDensity(temperature, density, fluid);
      entropy = GetEntropyFromTemperatureAndDensity(temperature, density, fluid);
    }

    #endregion

    #region 飽和状態の計算処理

    /// <summary>飽和圧力[kPa]から飽和物性値を計算する</summary>
    /// <param name="pressure">飽和圧力[kPa]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <param name="saturatedLiquidDensity">飽和液の密度[kg/m3]</param>
    /// <param name="saturatedVaporDensity">飽和蒸気の密度[kg/m3]</param>
    /// <param name="saturatedTemperature">飽和温度[K]</param>
    public static void GetSaturatedPropertyFromPressure
                (double pressure, Fluid fluid, out double saturatedLiquidDensity, out double saturatedVaporDensity, out double saturatedTemperature)
    {
      const double DELTA = 0.0001;
      const double ERR_TOL = 0.001;
      const int MAX_ITER = 10;

      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      //飽和温度の初期値を推定
      double pr = pressure / criticalPressure;
      double ts = cts[0];
      for (int i = 1; i < cts.Length; i++)
      {
        ts = ts * pr + cts[i];
      }

      double err = getGibbsEnergyDifference(ts, pressure, fluid, out saturatedLiquidDensity, out saturatedVaporDensity);
      int iter = 0;
      while (ERR_TOL < Math.Abs(err))
      {
        if (MAX_ITER < iter) throw new Exception("Iteration Error:GetSaturatedProperty");
        iter++;

        double err2 = getGibbsEnergyDifference(ts + DELTA, pressure, fluid, out saturatedLiquidDensity, out saturatedVaporDensity);
        double dedt = (err2 - err) / DELTA;
        ts -= err / dedt;
        err = getGibbsEnergyDifference(ts, pressure, fluid, out saturatedLiquidDensity, out saturatedVaporDensity);
      }

      saturatedTemperature = ts;
    }

    /// <summary>飽和温度[K]から飽和物性値を計算する</summary>
    /// <param name="temperature">飽和圧力[kPa]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <param name="saturatedLiquidDensity">飽和液の密度[kg/m3]</param>
    /// <param name="saturatedVaporDensity">飽和蒸気の密度[kg/m3]</param>
    /// <param name="saturatedPressure">飽和温度[kPa]</param>
    public static void GetSaturatedPropertyFromTemperature
        (double temperature, Fluid fluid, out double saturatedLiquidDensity, out double saturatedVaporDensity, out double saturatedPressure)
    {
      const double DELTA = 0.0001;
      const double ERR_TOL = 0.001;
      const int MAX_ITER = 10;

      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      //飽和圧力の初期値を推定
      double tr = temperature / criticalTemperature;
      double ps = cps[0];
      for (int i = 1; i < cps.Length; i++)
      {
        ps = ps * tr + cps[i];
      }

      double err = getGibbsEnergyDifference(temperature, ps, fluid, out saturatedLiquidDensity, out saturatedVaporDensity);
      int iter = 0;
      while (ERR_TOL < Math.Abs(err))
      {
        if (MAX_ITER < iter) throw new Exception("Iteration Error:GetSaturatedProperty");
        iter++;

        double err2 = getGibbsEnergyDifference(temperature, ps + DELTA, fluid, out saturatedLiquidDensity, out saturatedVaporDensity);
        double dedp = (err2 - err) / DELTA;
        ps -= err / dedp;
        err = getGibbsEnergyDifference(temperature, ps, fluid, out saturatedLiquidDensity, out saturatedVaporDensity);
      }

      saturatedPressure = ps;
    }

    /// <summary>温度[K]および圧力[kPa]から液体および気体のギブスエネルギーの差分[kJ]を計算する</summary>
    /// <param name="temperature">温度[K]</param>
    /// <param name="pressure">圧力[kPa]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <param name="liquidDensity">液体の密度[kg/m3]</param>
    /// <param name="vaporDensity">気体の密度[kg/m3]</param>
    /// <returns>ギブスエネルギーの差分[kJ]</returns>
    private static double getGibbsEnergyDifference
        (double temperature, double pressure, Fluid fluid, out double liquidDensity, out double vaporDensity)
    {
      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      double tr = temperature / criticalTemperature;

      //液体の密度
      //初期値をRackett式で設定
      double rc = criticalPressure / criticalDensity / criticalTemperature / gasConstant;
      double rhol = 1d / (Math.Pow(rc, 1d + Math.Pow(1d - tr, 2d / 7d)) / criticalPressure * gasConstant * criticalTemperature);
      getDensityFromPressureAndTemperature(pressure, temperature, fluid, ref rhol);
      liquidDensity = rhol;

      //気体の密度を推定
      //初期値を理想気体の状態式で設定
      double rhov = pressure / temperature / gasConstant;
      getDensityFromPressureAndTemperature(pressure, temperature, fluid, ref rhov);
      vaporDensity = rhov;

      //Gibbsエネルギーの差分を計算
      double gL = getResidualGibbsFreeEnergy(temperature, liquidDensity, fluid);
      double gV = getResidualGibbsFreeEnergy(temperature, vaporDensity, fluid);

      return gL - gV + gasConstant * temperature * Math.Log(liquidDensity / vaporDensity);
    }

    /// <summary>ニュートン法で圧力[kPa]および温度[K]から密度[kg/m3]を計算する</summary>
    /// <param name="pressure">圧力[kPa]</param>
    /// <param name="temperature">温度[K]</param>
    /// <param name="fluid">冷媒種類</param>
    public static double GetDensityFromPressureAndTemperature
        (double pressure, double temperature, Fluid fluid)
    {
      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      //密度の初期値を仮定
      double rho, rhoL, rhoV, satT;
      GetSaturatedPropertyFromPressure(pressure, fluid, out rhoL, out rhoV, out satT);
      if (satT < temperature) rho = rhoV;
      else rho = rhoL;

      getDensityFromPressureAndTemperature
          (pressure, temperature, fluid, ref rho);
      return rho;
    }

    /// <summary>ニュートン法で圧力[kPa]および温度[K]から密度[kg/m3]を計算する</summary>
    /// <param name="pressure">圧力[kPa]</param>
    /// <param name="temperature">温度[K]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <param name="density">密度[kg/m3]</param>
    private static void getDensityFromPressureAndTemperature
        (double pressure, double temperature, Fluid fluid, ref double density)
    {
      const double ERR_TOL = 0.0001;
      const int MAX_ITER = 10;

      double err = pressure - GetPressureFromTemperatureAndDensity(temperature, density, fluid);

      int iter = 0;
      while (ERR_TOL < Math.Abs(err))
      {
        if (MAX_ITER < iter) throw new Exception("Iteration Error:getDensityFromPressureAndTemperature");
        iter++;

        density += err / getDensityDerivativeOfPressure(temperature, density, fluid);
        err = pressure - GetPressureFromTemperatureAndDensity(temperature, density, fluid);
      }
    }

    #endregion

    #region Residual Property

    /// <summary>絶対温度[K]と密度[kg/m3]から残存比エンタルピー[kJ/kg]を計算する</summary>
    /// <param name="temperature">絶対温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>残存比エンタルピー[kJ/kg]</returns>
    private static double getResidualEnthalpy
        (double temperature, double density, Fluid fluid)
    {
      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      double tau = criticalTemperature / temperature;
      double rho = density / criticalDensity;

      double hr = 0;
      for (int m = alpha.GetLength(0) - 1; 0 <= m; m--)
      {
        double buff = 0;
        for (int n = alpha.GetLength(1) + 1; 2 <= n; n--)
        {
          buff = buff * rho + alpha[m, n - 2] * (n + m) / (n - 1d);
        }
        hr = hr * tau + buff;
      }
      hr *= rho;

      return hr / criticalDensity;
    }

    /// <summary>絶対温度[K]と密度[kg/m3]から残存比エントロピー[kJ/kgK]を計算する</summary>
    /// <param name="temperature">絶対温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>残存比エントロピー[kJ/kgK]</returns>
    private static double getResidualEntropy
        (double temperature, double density, Fluid fluid)
    {
      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      double tau = criticalTemperature / temperature;
      double rho = density / criticalDensity;

      double sr = 0;
      for (int m = alpha.GetLength(0) - 1; 1 <= m; m--)
      {
        double buff = 0;
        for (int n = alpha.GetLength(1) + 1; 2 <= n; n--)
        {
          buff = buff * rho + alpha[m, n - 2] * m / (n - 1d);
        }
        sr = sr * tau + buff;
      }
      sr *= rho * tau;

      return sr / (criticalDensity * temperature);
    }

    /// <summary>残存ギブスエネルギー[kJ]を計算する</summary>
    /// <param name="temperature">絶対温度[K]</param>
    /// <param name="density">密度[kg/m3]</param>
    /// <param name="fluid">冷媒種類</param>
    /// <returns>残存ギブスエネルギー[kJ]</returns>
    private static double getResidualGibbsFreeEnergy
        (double temperature, double density, Fluid fluid)
    {
      //必要に応じて冷媒初期化
      initRefrigerant(fluid);

      double tau = criticalTemperature / temperature;
      double rho = density / criticalDensity;

      double gr = 0;
      for (int m = alpha.GetLength(0) - 1; 0 <= m; m--)
      {
        double buff = 0;
        for (int n = alpha.GetLength(1) + 1; 2 <= n; n--)
        {
          buff = buff * rho + alpha[m, n - 2] * n / (n - 1d);
        }
        gr = gr * tau + buff;
      }
      gr *= rho;

      return gr / criticalDensity;
    }

    #endregion

    #region 理想気体の処理

    /// <summary>絶対温度[K]から理想気体の定圧比熱[kJ/kgK]を計算する</summary>
    /// <param name="temperature">温度[K]</param>
    /// <returns>理想気体の定圧比熱[kJ/kgK]</returns>
    private static double getIsobaricHeatCapacityOfIdealGas
      (double temperature)
    {
      double tr = temperature / criticalTemperature;
      double cp0 = ccp[ccp.Length - 1];
      for (int i = ccp.Length - 2; 0 <= i; i--)
      {
        cp0 = cp0 * tr + ccp[i];
      }
      return cp0 * gasConstant;
    }

    /// <summary>温度[K]から理想気体の定積比熱[kJ/kgK]を計算する</summary>
    /// <param name="temperature">絶対温度[K]</param>
    /// <returns>理想気体の定積比熱[kJ/kgK]</returns>
    private static double getIsovolumetricHeatCapacityOfIdealGas
      (double temperature)
    {
      return getIsobaricHeatCapacityOfIdealGas(temperature) - gasConstant;
    }

    /// <summary>温度[K]から理想気体の定圧比熱[kJ/kgK]の積分値を計算する</summary>
    /// <param name="temperature">温度[K]</param>
    /// <returns>理想気体の定圧比熱[kJ/kgK]の積分値</returns>
    private static double getIntegralCp0
        (double temperature)
    {
      double tr = temperature / criticalTemperature;
      double tr0 = refTemperature / criticalTemperature;
      double cp0 = 0;
      double cp0r = 0;
      for (int i = ccp.Length - 1; 0 <= i; i--)
      {
        cp0 = (cp0 + ccp[i] / (i + 1) * criticalTemperature) * tr;
        cp0r = (cp0r + ccp[i] / (i + 1) * criticalTemperature) * tr0;
      }
      return (cp0 - cp0r) * gasConstant;
    }

    /// <summary>温度[K]から理想気体の定圧比熱[kJ/kgK]/Tの積分値を計算する</summary>
    /// <param name="temperature">温度[K]</param>
    /// <returns>理想気体の定圧比熱[kJ/kgK]/Tの積分値</returns>
    private static double getIntegralCp0T
        (double temperature)
    {
      double tr = temperature / criticalTemperature;
      double tr0 = refTemperature / criticalTemperature;
      double cp0 = 0;
      double cp0r = 0;

      for (int i = ccp.Length - 1; 0 < i; i--)
      {
        cp0 = cp0 * tr + ccp[i] / i;
        cp0r = cp0r * tr0 + ccp[i] / i;
      }
      cp0 *= tr;
      cp0r *= tr0;

      return ((cp0 + ccp[0] * Math.Log(temperature)) -
          (cp0r + ccp[0] * Math.Log(refTemperature))) * gasConstant;
    }

    #endregion

  }
}
